/*
 * capture.c, vi: tabstop=4
 * capture module, controled by timer functions
 *
 * Copyright (C) 1997-98 Rasca, Berlin
 * EMail: thron@gmx.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <limits.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/XWDFile.h>
#include <X11/Xproto.h>
#include <X11/Xlibint.h>
#ifdef HasSharedMemory
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif
#ifdef HasDGA
#include <X11/extensions/xf86dga.h>
#endif
#ifdef HasVideo4Linux
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include "video.h"
#endif

#include "capture.h"
#include "job.h"
#include "control.h"
#include "main.h"
#include "script.h"


#define display() XtDisplay(job->toplevel)
#define screen() DefaultScreen(display())

/*
 * just read new data in the image structure, the image
 * structure inclusive the data area must be allocated before
 */
static Boolean
XGetZPixmap (Display *dpy, Drawable d, XImage *image, int x, int y)
{
    xGetImageReply rep;
    register xGetImageReq *req;
    long nbytes;

    if (!image)
		return (False);
    LockDisplay(dpy);
    GetReq (GetImage, req);
    /*
     * first set up the standard stuff in the request
     */
    req->drawable = d;
    req->x = x;
    req->y = y;
    req->width = image->width;
    req->height = image->height;
    req->planeMask = AllPlanes;
    req->format = ZPixmap;

    if (_XReply (dpy, (xReply *) &rep, 0, xFalse) == 0 ||
        rep.length == 0) {
        UnlockDisplay (dpy);
        SyncHandle ();
        return (False);
    }

    nbytes = (long)rep.length << 2;
	_XReadPad (dpy, image->data, nbytes);

    UnlockDisplay (dpy);
    SyncHandle();
	return (True);
}

/*
 * timer callback for capturing
 */
void
TCbCaptureX11 (XtPointer xtp, XtIntervalId *id)
{
	Job *job = (Job *) xtp;
	static char file[PATH_MAX+1];
	static XImage *image = NULL;
	static void *fp = NULL;
	long time;
	struct timeval curr_time;

        if ( job->script_file ) {
             run_script(job);
        }

#ifdef DEBUG2
	printf ("TCbCapture() pic_no=%d\n", job->pic_no);
#endif
	if ((job->state & VC_PAUSE) && !(job->state & VC_STEP)) {
		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
					job->time_per_frame,(XtTimerCallbackProc)job->capture, job);

	} else if (job->state & VC_REC) {
		if (job->max_frames && (
				(job->pic_no - job->start_no )> job->max_frames - 1)) {
			if (job->flags & FLG_RUN_VERBOSE)
				printf ("Stoped! pic_no=%d max_frames=%d\n",
					job->pic_no, job->max_frames);
			if (!(job->flags & FLG_RUN_VERBOSE))
				ChangeLabel (job->pic_no);
			goto CLEAN_X11;
		}
		job->state &= ~VC_STEP;
		gettimeofday (&curr_time, NULL);
		time = curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000;

		if ((!(job->flags & FLG_MULTI_IMAGE)) ||
			((job->flags & FLG_MULTI_IMAGE) && (job->state & VC_START))) {
			if (job->flags & FLG_MULTI_IMAGE)
				strcpy (file, job->file);
			else
				sprintf (file, job->file, job->pic_no);
			fp = (*job->open)(file, job->open_flags);
			if (!fp) {
				perror (file);
				job->state = VC_STOP;
				return;
			}
		}
		if (job->state & VC_START) {
			/*
			 * the first time this procedure is started
			 * we must create a new image ..
			 */
			image = XGetImage (display(), RootWindow(display(),
						DefaultScreen(display())), job->area->x, job->area->y,
						job->area->width, job->area->height, AllPlanes,ZPixmap);
			if (!image) {
				printf ("Can't get image: %dx%d+%d+%d\n", job->area->width,
					job->area->height, job->area->x, job->area->y);
					job->state = VC_STOP;
			} else {
				(*job->save) (fp, image, job);
				job->state &= ~VC_START;
			}
		} else {
			/* just read new data in the image structure
			 */
			if (XGetZPixmap (display(),
						RootWindow (display(), DefaultScreen(display())),
						image, job->area->x, job->area->y))
				(*job->save) (fp, image, job);
			else
				printf ("XGetZPixmap returned 'False'!\n");
		}
		if (!(job->flags & FLG_MULTI_IMAGE))
			(*job->close) (fp);

		/* substract the time we needed for creating and saving
		 * the frame to the file
		 */
		gettimeofday (&curr_time, NULL);
		time = (curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000) - time;
		time = job->time_per_frame - time;
		if (time < 0) {
			if (job->flags & FLG_RUN_VERBOSE)
			printf ("missing %ld milli secs (%d needed per frame), pic no %d\n",
					time, job->time_per_frame, job->pic_no);
			time = 0;
		}
		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
						time, (XtTimerCallbackProc)job->capture, job);
		job->pic_no += job->step;
		/* update the label if we have time to do this */
		if (time)
			ChangeLabel (job->pic_no);
	} else {
		/* clean up */
		CLEAN_X11:
		/* may be the last update failed .. so do it here before stop
		 */
		ChangeLabel (job->pic_no);
		job->state = VC_STOP;
		if (image) {
			XDestroyImage (image);
			image = NULL;
		}
		/* set the sensitive stuff for the control panel
		 */
		CbStop (NULL, NULL, NULL);

		/* clean up the save routines in xtoXXX.c
		 */
		if (job->clean)
			(*job->clean) (job);
		if (job->flags & FLG_MULTI_IMAGE)
			if (fp)
				(*job->close) (fp);
		fp = NULL;
	}
}


#ifdef HasSharedMemory
/*
 * timer callback for capturing with shared memory
 */
void
TCbCaptureSHM (XtPointer xtp, XtIntervalId *id)
{
	Job *job = (Job *) xtp;
	static char file[PATH_MAX+1];
	static XImage *image = NULL;
	static XShmSegmentInfo shminfo;
	static void *fp = NULL;
	Visual *visual = job->win_attr.visual;
	unsigned int depth = job->win_attr.depth;
	long time;
	struct timeval curr_time;

        if ( job->script_file ) {
             run_script(job);
        }

#ifdef DEBUG2
	printf ("TCbCaptureSHM() pic_no=%d flags=%d\n", job->pic_no, job->flags);
#endif
	if ((job->state & VC_PAUSE) && !(job->state & VC_STEP)) {
		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
					job->time_per_frame,(XtTimerCallbackProc)job->capture, job);
	} else if (job->state & VC_REC) {
		if (job->max_frames && (
				(job->pic_no - job->start_no )> job->max_frames - 1)) {
			if (job->flags & FLG_RUN_VERBOSE)
				printf ("Stoped! pic_no=%d max_frames=%d\n",
					job->pic_no, job->max_frames);
			if (!(job->flags & FLG_RUN_VERBOSE))
				ChangeLabel (job->pic_no);
			goto CLEAN_SHM;
		}
		job->state &= ~VC_STEP;
		gettimeofday (&curr_time, NULL);
		time = curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000;

		if ((!(job->flags & FLG_MULTI_IMAGE)) ||
			((job->flags & FLG_MULTI_IMAGE) && (job->state & VC_START))) {

			if (job->flags & FLG_MULTI_IMAGE)
				strcpy (file, job->file);
			else
				sprintf (file, job->file, job->pic_no);
			fp = (*job->open)(file, job->open_flags);
			if (!fp) {
				perror (file);
				job->state = VC_STOP;
				return;
			}
		}
		if (job->state & VC_START) {
			/*
			 * the first time this procedure is called so
			 * we must create a new ximage ..
			 */
			image = XShmCreateImage (display(), visual, depth, ZPixmap, NULL,
						&shminfo, job->area->width, job->area->height);
			if (!image) {
				printf ("Can't get image: %dx%d+%d+%d\n", job->area->width,
					job->area->height, job->area->x, job->area->y);
				job->state = VC_STOP;
				CbStop (NULL, NULL, NULL);
				return;
			}
			shminfo.shmid = shmget (IPC_PRIVATE,
					image->bytes_per_line * image->height, IPC_CREAT|0777);
			if (shminfo.shmid == -1) {
				printf ("Fatal: Can't get shared memory!\n");
				exit (1);
			}
			shminfo.shmaddr = image->data = shmat (shminfo.shmid, 0, 0);
			shminfo.readOnly = False;

			if (XShmAttach (display(), &shminfo) == 0) {
				printf ("Fatal: Failed to attach shared memory!\n");
				/* needs some better error subroutine :) */
				exit(1);
			}
		}
		/* just read new data in the image structure
		 */
		if (XShmGetImage (display(),
						RootWindow (display(), DefaultScreen(display())),
						image, job->area->x, job->area->y, AllPlanes)) {
			(*job->save) (fp, image, job);
			job->state &= ~VC_START;
		} else {
				printf ("XShmGetImage() returned 'False'!\n");
		}
		if (!(job->flags & FLG_MULTI_IMAGE))
			(*job->close) (fp);
		else if (job->flags & FLG_SYNC) {
			if (job->open == (void *(*)(char *,char*))fopen)
				fdatasync (fileno(fp));
		}

		/* substract the time we needed for creating and saving
		 * the frame to the file
		 */
		gettimeofday (&curr_time, NULL);
		time = job->time_per_frame
				- ((curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000) - time);
		if (time < 0) {
			if (job->flags & FLG_RUN_VERBOSE)
				printf ("missing %ld milli secs, %d needed /p frame, no. %d\n",
					- time, job->time_per_frame, job->pic_no);
			time = 0;
		}

		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
						time, (XtTimerCallbackProc)job->capture, job);
		job->pic_no += job->step;
		if (time)
			ChangeLabel (job->pic_no);
	} else {
		/* clean up */
		CLEAN_SHM:
		ChangeLabel (job->pic_no);
		job->state = VC_STOP;
		if (image) {
			XShmDetach (display(), &shminfo);
			image->data = NULL;
			XDestroyImage (image);
			image = NULL;
			shmdt (shminfo.shmaddr);
			shmctl (shminfo.shmid, IPC_RMID, 0);
		}
		CbStop (NULL, NULL, NULL);
		if (job->clean) {
			(*job->clean) (job);
		}
		if (job->flags & FLG_MULTI_IMAGE)
			if (fp)
				(*job->close) (fp); 
		fp = NULL;
	}
}

#endif /* HasSharedMemory */

#ifdef HasVideo4Linux
/*
 * timer callback for capturing direct from bttv driver (only linux)
 */
#ifndef linux
#error only for linux
#endif
/* from bttv.h */

void
TCbCaptureV4L (XtPointer xtp, XtIntervalId id *)
{
	Job *job = (Job *) xtp;
	static char file[PATH_MAX+1];
	static XImage *image = NULL;
	static void *fp = NULL;
	static VIDEO *video = 0;
	static int size;
	static struct video_mmap	vi_mmap;
	static struct video_mbuf	vi_memb;
	static struct video_picture	vi_pict;
	long time;
	struct timeval curr_time;

        if ( job->script_file ) {
             run_script(job);
        }

#ifdef DEBUG2
	printf ("TCbCaptureV4L() pic_no=%d\n", job->pic_no);
#endif
	if ((job->state & VC_PAUSE) && !(job->state & VC_STEP)) {
		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
					job->time_per_frame,(XtTimerCallbackProc)job->capture, job);

	} else if (job->state & VC_REC) {
		if (job->max_frames && (
				(job->pic_no - job->start_no )> job->max_frames - 1)) {
			if (job->flags & FLG_RUN_VERBOSE)
				printf ("Stoped! pic_no=%d max_frames=%d\n",
					job->pic_no, job->max_frames);
			if (!(job->flags & FLG_RUN_VERBOSE))
				ChangeLabel (job->pic_no);
			goto CLEAN_V4L;
		}
		job->state &= ~VC_STEP;
		gettimeofday (&curr_time, NULL);
		time = curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000;

		if ((!(job->flags & FLG_MULTI_IMAGE)) ||
			((job->flags & FLG_MULTI_IMAGE) && (job->state & VC_START))) {
			if (job->flags & FLG_MULTI_IMAGE)
				strcpy (file, job->file);
			else
				sprintf (file, job->file, job->pic_no);
			fp = (*job->open)(file, job->open_flags);
			if (!fp) {
				perror (file);
				job->state = VC_STOP;
				return;
			}
		}
		if (job->state & VC_START) {
			/*
			 * the first time this procedure is started
			 * we must prepare some stuff ..
			 */
			if (!job->bpp) /* use depth of the default window */
				job->bpp = job->win_attr.depth;

			sync(); /* remove if your bttv driver runs stable .. */
			video = video_open (job->video_dev, O_RDWR);
			if (!video) {
				perror (job->video_dev);
				goto CLEAN_V4L;
			}
			vi_pict.depth = 0;
			/* read default values for hue, etc.. */
			ioctl (video->fd, VIDIOCGPICT, &vi_pict);

			printf ("%d->%d %d\n", job->bpp, vi_pict.depth, vi_pict.palette);
			switch (job->bpp) {
				case 24:
					vi_mmap.format = vi_pict.palette = VIDEO_PALETTE_RGB24;
					break;
				case 16:
					vi_mmap.format = vi_pict.palette = VIDEO_PALETTE_RGB565;
					break;
				case 15:
					vi_mmap.format = vi_pict.palette = VIDEO_PALETTE_RGB555;
					break;
				default:
					printf ("Fatal: unsupported bpp (%d)\n", job->bpp);
					exit (3);
					break;
			}
			ioctl (video->fd, VIDIOCSPICT, &vi_pict);
			printf ("%d->%d %d\n", job->bpp, vi_pict.depth, vi_pict.palette);


			vi_memb.size = 0;
			ioctl(video->fd, VIDIOCGMBUF, &vi_memb);
			printf ("%d %d %d\n", vi_memb.size, vi_memb.frames,vi_memb.offsets);

			image = (XImage *)XtMalloc (sizeof(XImage));
			if (!image) {
				printf ("Can't get image: %dx%d+%d+%d\n", job->area->width,
					job->area->height, job->area->x, job->area->y);
					goto CLEAN_V4L;
			}
			switch (job->bpp) {
				case 24:
					image->red_mask = 0xFF0000;
					image->green_mask = 0x00FF00;
					image->blue_mask = 0x0000FF;
					break;
				case 16:
					image->red_mask = 0x00F800;
					image->green_mask = 0x0007E0;
					image->blue_mask = 0x00001F;
					break;
				case 15:
					image->red_mask = 0x00F800;
					image->green_mask = 0x0007E0;
					image->blue_mask = 0x00001F;
					break;
				default:
					printf ("Fatal: unsupported bpp (%d)\n", job->bpp);
					exit (3);
					break;
			}
			image->width = job->area->width;
			image->height= job->area->height;
			image->bits_per_pixel = job->bpp;
			image->bytes_per_line = job->bpp/8 * image->width;
			image->byte_order = MSBFirst;
			size = image->width * image->height * job->bpp;
			video->size = vi_memb.size;
			video_mmap (video, 1);
			if (video->mmap == NULL) {
				perror ("mmap()");
				goto CLEAN_V4L;
			}

			vi_mmap.width = image->width;
			vi_mmap.height= image->height;
			vi_mmap.frame = 0;
			image->data = video->mmap;
		}
		/* just read new data in the image structure
		 */
		if (ioctl (video->fd, VIDIOCMCAPTURE, &vi_mmap) < 0) {
			perror ("ioctl(capture)");
			/* if (vb.frame) vb.frame = 0; else vb.frame = 1; */
			goto CLEAN_V4L;
		}
		printf ("syncing ..\n");
		if (ioctl (video->fd, VIDIOCSYNC, vi_mmap.frame) < 0) {
			perror ("ioctl(sync)");
		}
		printf ("synced()\n");

		(*job->save) (fp, image, job);
		job->state &= ~VC_START;

		if (!(job->flags & FLG_MULTI_IMAGE))
			(*job->close) (fp);

		/* substract the time we needed for creating and saving
		 * the frame to the file
		 */
		gettimeofday (&curr_time, NULL);
		time = (curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000) - time;
		time = job->time_per_frame - time;
		if (time < 0) {
			if (job->flags & FLG_RUN_VERBOSE)
			printf ("missing %ld milli secs (%d needed per frame), pic no %d\n",
					time, job->time_per_frame, job->pic_no);
			time = 0;
		}
		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
						time, (XtTimerCallbackProc)job->capture, job);
		job->pic_no += job->step;
		/* update the label if we have time to do this */
		if (time)
			ChangeLabel (job->pic_no);
	} else {
		/* clean up */
		CLEAN_V4L:
		/* may be the last update failed .. so do it here before stop
		 */
		ChangeLabel (job->pic_no);
		job->state = VC_STOP;
		if (image) {
			XtFree ((char*)image);
			image = NULL;
		}
		if (video) {
			video_mmap (video, 0);
			video_close (video);
			video = NULL;
		}
		/* set the sensitive stuff for the control panel
		 */
		CbStop (NULL, NULL, NULL);

		/* clean up the save routines in xtoXXX.c
		 */
		if (job->clean)
			(*job->clean) (job);
		if (job->flags & FLG_MULTI_IMAGE)
			if (fp)
				(*job->close) (fp);
		fp = NULL;
	}
}
#endif /* HasVideo4Linux */

#ifdef HasDGA
/*
 * direct graphic access
 * this doesn't work until now and may be removed in future..!?
 */
void
TCbCaptureDGA (XtPointer xtp, XtIntervalId id *)
{
	Job *job = (Job *) xtp;
	static char file[PATH_MAX+1];
	static XImage *image = NULL;
	static void *fp;
	static int size;
	static Display *dpy;
	long time;
	struct timeval curr_time;

        if ( job->script_file ) {
             run_script(job);
        }

#ifdef DEBUG2
	printf ("TCbCaptureDGA() pic_no=%d state=%d\n", job->pic_no, job->state);
#endif
	if ((job->state & VC_PAUSE) && !(job->state & VC_STEP)) {
		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
					job->time_per_frame,(XtTimerCallbackProc)job->capture, job);

	} else if (job->state & VC_REC) {
		if (job->max_frames && (
				(job->pic_no - job->start_no )> job->max_frames - 1)) {
			if (job->flags & FLG_RUN_VERBOSE)
				printf ("Stoped! pic_no=%d max_frames=%d\n",
					job->pic_no, job->max_frames);
			if (!(job->flags & FLG_RUN_VERBOSE))
				ChangeLabel (job->pic_no);
			goto CLEAN_DGA;
		}
		job->state &= ~VC_STEP;
		gettimeofday (&curr_time, NULL);
		time = curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000;

		if ((!(job->flags & FLG_MULTI_IMAGE)) ||
			((job->flags & FLG_MULTI_IMAGE) && (job->state & VC_START))) {
			if (job->flags & FLG_MULTI_IMAGE)
				strcpy (file, job->file);
			else
				sprintf (file, job->file, job->pic_no);
			fp = (*job->open)(file, job->open_flags);
			if (!fp) {
				perror (file);
				job->state = VC_STOP;
				return;
			}
		}
		if (job->state & VC_START) {
			/*
			 * the first time this procedure is started
			 * we must create a new image structure ..
			 */
			dpy = XtDisplay (job->toplevel);
			image = (XImage *)XtMalloc (sizeof(XImage));
			if (!image) {
				printf ("Can't get image: %dx%d+%d+%d\n", job->area->width,
					job->area->height, job->area->x, job->area->y);
					goto CLEAN_DGA;
			}
			image->width = job->area->width;
			image->height= job->area->height;
			image->bits_per_pixel = 3 * 8;
			image->bytes_per_line = 3 * image->width;
			image->byte_order = MSBFirst;
			size = image->width * image->height;
			{ int width, bank, ram;
				char *base;
			XF86DGAGetVideo (dpy, XDefaultScreen (dpy), (char **) &base,
				&width, &bank, &ram);
				image->data = base;
			}
			XF86DGADirectVideo (dpy, XDefaultScreen(dpy),XF86DGADirectGraphics);
			(*job->save) (fp, image, job);
			XF86DGADirectVideo (dpy, XDefaultScreen(dpy), 0);
			job->state &= ~VC_START;
		} else {
		/* just read new data in the image structure
		 */
			XF86DGADirectVideo (dpy, XDefaultScreen(dpy),XF86DGADirectGraphics);
			(*job->save) (fp, image, job);
			XF86DGADirectVideo (dpy, XDefaultScreen(dpy), 0);
		}
		if (!(job->flags & FLG_MULTI_IMAGE))
			(*job->close) (fp);

		/* substract the time we needed for creating and saving
		 * the frame to the file
		 */
		gettimeofday (&curr_time, NULL);
		time = (curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000) - time;
		time = job->time_per_frame - time;
		if (time < 0) {
			if (job->flags & FLG_RUN_VERBOSE)
			printf ("missing %ld milli secs (%d needed per frame), pic no %d\n",
					time, job->time_per_frame, job->pic_no);
			time = 0;
		}
		XtAppAddTimeOut (XtWidgetToApplicationContext(job->toplevel),
						time, (XtTimerCallbackProc)job->capture, job);
		job->pic_no += job->step;
		/* update the label if we have time to do this */
		if (time)
			ChangeLabel (job->pic_no);
	} else {
		/* clean up */
		CLEAN_DGA:
		/* may be the last update failed .. so do it here before stop
		 */
		ChangeLabel (job->pic_no);
		job->state = VC_STOP;
		if (image) {
			XtFree ((char*)image);
			image = NULL;
		}
		XF86DGADirectVideo (dpy, XDefaultScreen(dpy), 0);
		/* set the sensitive stuff for the control panel
		 */
		CbStop (NULL, NULL, NULL);

		/* clean up the save routines in xtoXXX.c
		 */
		if (job->clean)
			(*job->clean) (job);
		if (job->flags & FLG_MULTI_IMAGE)
			if (fp)
				(*job->close) (fp);
		fp = NULL;
	}
}
#endif /* HasDGA */
