/*
 * job.c
 *
 * Copyright (C) 1997,98 Rasca, Berlin
 * EMail: thron@gmx.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Intrinsic.h>
#ifdef HasZlib
#include <zlib.h>
#endif
#include "job.h"
#include "capture.h"
#include "main.h"
#include "util.h"
#include "sound.h"
#include "xtopnm.h"
#include "xtoxwd.h"
#include "xtopng.h"
#include "xtomng.h"
#include "xtojpg.h"
#include "xtoqtf.h"
#include "script.h"

static Job *job;

/*
 */
Job *
job_init (int flags, Widget toplevel)
{
	job = (Job *) malloc (sizeof (Job));
	if (!job) {
		perror ("job_init()");
		exit (1);
	}
	job->flags = flags;
	job->file = NULL;
	job->step = 1;
	job->target = CAP_XWD;
	job->state = VC_STOP;
	job->toplevel = toplevel;
	job->color_table = NULL;
	job->get_colors = (void*(*)(XColor *, int))NULL;
	job->capture = TCbCaptureX11;
	job->clean = (void(*)(Job *))NULL;
	job->colors = NULL;
        job->script_file = NULL;
#ifdef DEBUG2
	printf ("job_init (flags=%d)\n", flags);
#endif
	return (job);
}

/*
 */
Job *
job_ptr (void)
{
	return (job);
}


/*
 */
void
job_set_fps (float fps)
{
	if (fps <= 0)
		fps = 0.00001; /* min value */
	job->time_per_frame = 1000 / fps;
	job->fps = fps;
}

/*
 * return frames per second
 */
float
job_fps (void)
{
	return (job->fps);
}

/*
 * set and check some parameters for the sound device
 */
int
job_set_sound_dev (char *snd, int rate, int size)
{
#ifdef HasOSS
	if (!snd_init (snd)) {
		job->snd_rate = 0;
		job->snd_smplsize = 0;
		return (0);
	}
	job->snd_device = strdup (snd);
	if (snd_open_rec ()) {
		job->snd_rate = snd_set_speed (rate);
		job->snd_smplsize = snd_set_samplesize (size);
		job->snd_stereo = snd_set_stereo (1);
	}
#endif
	return (1);
}

/*
 */
void
job_set_file (char *file)
{
	if (!file)
		return;
	if (strstr (file, ".ppm") ||
		strstr (file, ".pnm") ||
		strstr (file, ".pgm") ) {
		job->target = CAP_PNM;
		job->get_colors = PPMcolorTable;
		job->clean = PnmClean;
	} else if (strstr (file, ".png")) {
#ifdef HasPNG
		job->target = CAP_PNG;
		job->get_colors = PNGcolorTable;
		job->clean = PngClean;
	} else if (strstr (file, ".mng")) {
		job->target = CAP_MNG;
		job->get_colors = PNGcolorTable;
		job->flags |= FLG_MULTI_IMAGE;
		job->clean = MngClean;
#else
		printf ("PNG not supported!\n");
		exit (1);
#endif
	} else if (strstr (file, ".jpg") ||
				strstr (file, ".jpeg")) {
#ifdef HasJPEG
		job->target = CAP_JPG;
		job->get_colors = JPGcolorTable;
		job->clean = JpgClean;
#else
		printf ("JPEG not supported!\n");
		exit (1);
#endif
	} else if (strstr (file, ".mov")) {
		job->target = CAP_QTR;
		job->clean = QtClean;
		job->flags |= FLG_MULTI_IMAGE;
#ifndef HasJPEG
		job->quality = 100;
#endif
#ifndef HasPNG
		job->compress= 0;
#endif
		if (job->quality && (job->quality < 100)) {
			/* jpeg codec */
			job->target = CAP_QTJ;
			job->compress = 0;
		} else if (job->compress) {
			/* png codec */
			job->target = CAP_QTP;
		}
	} else {
		/* default is XWD
		 */
		job->target = CAP_XWD;
		job->get_colors = XWDcolorTable;
	}

#ifdef HasZlib
	if (job->compress && !(
			(job->target == CAP_PNG) || (job->target == CAP_MNG) ||
			(job->target == CAP_QTP))) {
		int len = strlen (file);

		job->open = (void*(*)(char*,char*))gzopen;
		job->close= (int(*)(void*))gzclose;
		job->write = zwrite;
		if (job->file && (job->file != file))
			free (job->file);
		if (!strstr (file, ".gz")) {
			job->file = (char *) malloc (len + 4 + 1);
			sprintf (job->file, "%s.gz", file);
		} else {
			job->file = strdup (file);
		}
		sprintf (job->open_flags, "wb%d", job->compress);
	} else
#endif
	{
		job->open = (void*(*)(char*,char*))fopen;
		job->close= (int(*)(void*))fclose;
		if (job->target == CAP_QTJ) {
		 job->write = (int(*)(void*, unsigned int, unsigned int, void *))jwrite;
		} else if (job->target == CAP_QTP) {
		 job->write = (int(*)(void*, unsigned int, unsigned int, void *))pwrite;
		} else {
		 job->write = (int(*)(void*, unsigned int, unsigned int, void *))fwrite;
		}
		job->file = strdup (file);
		if (strstr (file, ".gz")) {
			int len = strlen (file);
			job->file[len-3] = '\0';
		}
		sprintf (job->open_flags, "wb");
	}
#ifdef HasBTTV
	if (job->flags & FLG_USE_V4L) {
		job->capture = TCbCaptureV4L;
	}
#endif
}

/*
 */
char *
job_file (void)
{
	return (job->file);
}

void
job_set_script_file (char* script_file)
{
     job->script_file  = script_file;
     job->from_script  = NULL;
     if ( script_file ) {
          job->script_state = SCRIPT_PRESENT;
     } else {
          job->script_state = 0;
     }
}

void
job_set_result_file (char* result_file)
{
     job->result_file = result_file;
     job->to_result   = NULL;
}

void
job_set_windowid (Window windowid)
{
     job->windowid = windowid;
     if ( windowid ) {
          adjust_grabbed_window(windowid);
     }
}

/*
 */
void
job_set_compression (int comp)
{
	if (comp < 0)
		comp = 0;
	else if (comp > 9)
		comp = 9;
	job->compress = comp;
}

/*
 */
int
job_compression (void)
{
	return (job->compress);
}

/*
 * find the correct capture function
 */
void
job_set_capture (void)
{
	int input = job->flags & FLG_SOURCE;

	switch (input) {
		case FLG_USE_SHM:
			job->capture = TCbCaptureSHM;
			break;
		case FLG_USE_DGA:
			job->capture = TCbCaptureDGA;
			break;
		case FLG_USE_V4L:
			job->capture = TCbCaptureV4L;
			break;
		default:
			job->capture = TCbCaptureX11;
			break;
	}
}

/*
 * return if shared memory is used or not
 */
int
job_shm (void)
{
	return (job->flags & FLG_USE_SHM);
}

/*
 */
void
job_set_quality (int quality)
{
	if (quality < 1)
		quality = 1;
	else if (quality > 100)
		quality = 100;
	job->quality = quality;
}

/*
 */
int job_quality (void)
{
	return (job->quality);
}

