/* $Id: script.c,v 1.6.1.4 2001/06/14 15:26:20 queinnec Exp $ */

#include <stdio.h>
#include <limits.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/XWDFile.h>
#include <X11/Xproto.h>
#include <X11/Xlibint.h>
#ifdef HasSharedMemory
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif
#ifdef HasDGA
#include <X11/extensions/xf86dga.h>
#endif
#ifdef HasVideo4Linux
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include "video.h"
#endif

#include "capture.h"
#include "job.h"
#include "control.h"
#include "script.h"
#include <X11/keysym.h>

/* 
   Return the time spent since the beginning of the script.
   This time is expressed in milli-seconds. 
*/

static unsigned long
get_relative_time (Job *job)
{
     unsigned long time;
     struct timeval curr_time;

     gettimeofday (&curr_time, NULL);
     time = curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000;
     return time - job->script_start_time;
}

/* Sleep for a number of milliseconds */

static void
sleep_milliseconds (unsigned int n)
{
     usleep(n * 1000UL);
}

/* 
   Send a character to the target window. 
   Many thanks to cl@agraf.liafa.jussieu.fr (Choun Tong LIEU). 
   As he mentioned, xterm does not accept characters sent like this,
   use rxvt instead! Alas, newlines do not seem to be sent ???
 */

static void
send_event (Job *job, XKeyEvent *pevmt)
{
     Display *dpy = XtDisplay(job->toplevel);
     Status s;

     if ( pevmt->keycode == 0 ) {
          /* Never send null events said Matthew Flatt <mflatt@cs.utah.edu> */
          return;
     };
     pevmt->type = KeyPress;
     s = XSendEvent(dpy, job->windowid, False, 0, (XEvent *)pevmt);
     if ( s == BadValue ) {
          printf("Problem sending character.\n");
          exit(19);
     } else if ( s == BadWindow ) {
          printf("Problem with grabbed window.\n");
          exit(19);
     }
     pevmt->type = KeyRelease;
     s = XSendEvent(dpy, job->windowid, False, 0, (XEvent *)pevmt);
     if ( s == BadValue ) {
          printf("Problem sending character.\n");
          exit(19);
     } else if ( s == BadWindow ) {
          printf("Problem with grabbed window.\n");
          exit(19);
     }
     /* XFlush(dpy); */
}

static void 
send_character (Job *job, char c, unsigned long mask)
{
     XKeyPressedEvent evmt;
     Display *dpy = XtDisplay(job->toplevel);
     Status s;
     int i, keycode;

#ifdef DEBUG
     printf("send_character: send character %c\n", c);
#endif

     evmt.display     = XtDisplay(job->toplevel);
     evmt.send_event  = False;
     evmt.window      = job->windowid;
     evmt.time        = CurrentTime;
     evmt.same_screen = True;
     evmt.state       = mask;
     switch (c) {
          case '\r': {
               evmt.keycode = XKeysymToKeycode(dpy, XK_Return);
               send_event(job, &evmt);
               break;
          }
          case '\n': {
               evmt.keycode = XKeysymToKeycode(dpy, XK_Return);
               send_event(job, &evmt);
               evmt.keycode = XKeysymToKeycode(dpy, XK_Linefeed);
               send_event(job, &evmt);
               break;
          }
          case '\t': {
               evmt.keycode = XKeysymToKeycode(dpy, XK_Tab);
               send_event(job, &evmt);
               break;
          }
          default: {
               /* Inspired by wool_send_key_to_client() from wl_event.c from
                  gwm-1.8d by Colas Nahaboo. */
               keycode = XKeysymToKeycode(dpy, c);
               if ( keycode == 0 ) {
                    printf("send_character: Unknown keysym 0x%x\n", c);
                    exit(25);
               };
               evmt.keycode = keycode;
               if ( c != XKeycodeToKeysym(dpy, keycode, 0) ) {
                    for (i = 0 ; i < 14 ; ++i) {
                         if ( c == XKeycodeToKeysym(dpy, keycode, 1<<i) ) {
                              evmt.state |= 1<<i;
                         };
                    }
               }
#ifdef DEBUG
               printf("send_character: mask is %8x\n", evmt.state);
#endif
               send_event(job, &evmt);
               break;
          }
     }
#ifdef DEBUG
     printf("send_character: character sent: %c\n", c);
#endif
}

/* 
   Finish the script.
*/

static void
terminate_script (Job *job)
{
     fclose(job->to_result);
     fclose(job->from_script);
     job->to_result = NULL;
     job->from_script = NULL;
     /* the script is not present any longer: */
     job->script_state = SCRIPT_DONE;
     /* ensure capture to end. */
     job->state &= ~(VC_PAUSE | VC_REC);
}

static void
process_next_token (Job *job);

/* 
   Process a non-empty @command.
*/

static void
process_command (Job *job, char *command)
{
     unsigned int n;
     unsigned long time;

#ifdef DEBUG
     printf("run_script: process command: %s\n", command);
#endif

     if ( strlen(command) == 0 ) {
          /* Ignore that command and process next one. */
          return;

#define CompareWith(s) \
  ( 0 == strncmp(command, s, sizeof(s)-1) )

     } else if ( CompareWith("#") ) {
          /* Nothing, this is a comment and process next one. */

     } else if ( CompareWith("capture") ) {
          /* Capture frame now. */
          return;

     } else if ( CompareWith("sleep") ) {
          /* sleep that number of milliseconds. */
          command += strlen("sleep");
          n = (unsigned int) strtol(command, 0, 0);
          sleep_milliseconds(n);

     } else if ( CompareWith("sh") ) {
          /* Run a shell command for its effect. (popen is another
             similar command whose output is piped towards the
             monitored window). */
          command += strlen("sh");
          system(command);

     } else if ( CompareWith("echo") ) {
          /* Output a string into the result file. This is mainly used
             to specify the sound files to start with some screens. */
          command += strlen("echo");
          while ( isspace(*command) ) {
               command++;
          }
          time = get_relative_time(job);
          fprintf(job->to_result, "%s at %lu\n", command, time);

     } else if ( CompareWith("subtitle") ) {
          /* Specify the accompanying subtitle. This text will be
             output into the result file so it can be displayed by the
             ultimate player. */
          command += strlen("subtitle");
          while ( isspace(*command) ) {
               command++;
          }
          time = get_relative_time(job);
          fprintf(job->to_result, "subtitle \"%s\" at %lu\n", command, time);

     } else if ( CompareWith("sync") ) {
          /* Wait until a number of milliseconds is exhausted from the
             beginning of the script. This is useful if one needs to
             resynchronize the script with some external event.  In some
             future, the argument may be <name>+milliseconds where <name>
             is the name of a mark (see @mark).  */
          command += strlen("sync");
          n = (unsigned int) strtol(command, 0, 0);
          time = get_relative_time(job);
          if ( n >= time ) {
               sleep_milliseconds((n - time));
          } else {
               printf("run_script: cannot synchronize with past.\n");
               exit(17);
          }

     } else if ( CompareWith("fps") ) {
          /* This allows to set the time lag between characters of the
             script. The number is expressed as a float specifying the
             number of frames per second.  */
          command += strlen("fps");
          job_set_fps((float) strtod(command, 0));

     } else if ( CompareWith("quit") ) {
          /* Capture the last frame then exit. */
          job->script_state |= SCRIPT_STOP;

     } else if ( CompareWith("\\") ) {
          /* Send a character with some modifier.  Legal characters
             are for instance, \M-a or \C-b or \M-\C-c or \C-\M-d */
          unsigned long int mask = 0;

          while ( 1 ) {
               if ( CompareWith("\\M-") ) {
                    /* Meta or alt */
                    command += strlen("\\M-");
                    mask |= Mod1Mask;
               } else if ( CompareWith("\\C-") ) {
                    /* Control */
                    command += strlen("\\C-");
                    mask |= ControlMask;
               } else if ( CompareWith("\\S-") ) {
                    /* Shift */
                    command += strlen("\\S-");
                    mask |= ShiftMask;
               } else if ( CompareWith("\\") ) {
                    command += strlen("\\");
                    send_character(job, *command, mask);
                    break;
               } else {
                    send_character(job, *command, mask);
                    break;
               }
          }
          

/* ******************************************** Not yet implemented
     } else if ( CompareWith("popen") ) {
     } else if ( CompareWith("include") ) {
     } else if ( CompareWith("mark") ) {
*********************************************** */
          
     } else {
          printf("run_script: unrecognized command %s\n", command);
          exit(7);      /* TO BE FIXED */
     }
     /* Process next command if any */
     process_next_token(job);
}

/* 
   Process the next character or @command of the script file
   then capture the current frame.
*/

static void
process_next_token (Job *job)
{
     int c;

     if ( job->script_state & SCRIPT_STOP ) {
          job->script_state = SCRIPT_DONE;
          return;
     } else {
          c = fgetc(job->from_script);
     }

#ifdef DEBUG
     if ( c == EOF ) {
          printf("run_script: process EOF\n");
     } else {
          printf("run_script: process character %c\n", c);
     }
#endif

     if ( c == EOF ) {
          terminate_script(job);
     } else {
          if ( c == '\\' ) {
               c = fgetc(job->from_script);
               if ( c != EOF ) {
                    send_character(job, c, 0);
               } else {
                    printf("Script file ends with a backslash.\n");
                    exit(5);   /* TO BE FIXED */
               }
          } else if ( c == '@' ) {
#define COMMAND_LENGTH  1024
               char command[COMMAND_LENGTH] = { '\0', '\0' };
               char *commandp = &command[0];

               /* Accumulate all character until the end of the line in
                  the command buffer. */
               while ( (c = fgetc(job->from_script)) != EOF ) {
                    *commandp++ = c;
                    if ( c == '\n' ) {
                         commandp--;
                         break;
                    } else if ( (commandp-&command[0]) >= COMMAND_LENGTH ) {
                         /* The buffer is full! */
                         *commandp = '\0';
                         printf("run_script: oversized command: %s\n", command);
                         exit(23);
                    };
               }
               /* non empty commands are processed. */
               if ( command != commandp ) {
                    *commandp = '\0';
                    process_command(job, command);
               };
          } else {
               send_character(job, c, 0);
          }
     }

     /* Capture frame now! */
}

/* 
   Process the script file if any.
   This function is regularly called by a TcbCaptureXXX call-back (see
   capture.c). 
*/

void
run_script (Job *job)
{

#ifdef DEBUG
     printf("xscript: $Id: script.c,v 1.6.1.4 2001/06/14 15:26:20 queinnec Exp $\n");
#endif

     if ( !(job->script_state & SCRIPT_PRESENT) ) {
          return;
     }

     /* Open the script file if not yet done. */
     if ( !job->from_script ) {

#ifdef DEBUG
          printf("run_script: open script file %s\n", job->script_file);
#endif
          job->from_script = fopen(job->script_file, "r");
          if ( !job->from_script ) {
               printf("Cannot open script file %s\n", job->script_file);
               exit(3);                  /* TO BE FIXED */
          }
     }

     /* Compute the name of the result file if not yet done. */
     if ( !job->result_file ) {
          size_t nfnsize = strlen(EXTENSION) + strlen(job->script_file);
          job->result_file = malloc(nfnsize + 1);
          sprintf(job->result_file, "%s.t", job->script_file);
     }

     /* Open the result file if not yet done and start processing the
        script file.
     */
     if ( !job->to_result ) {
          unsigned long time;
          struct timeval curr_time;

#ifdef DEBUG
          printf("run_script: open result file %s\n", job->result_file);
#endif
          job->to_result = fopen(job->result_file, "w");
          if ( !job->to_result ) {
               printf("Cannot open result file %s\n", job->result_file);
               exit(11);                  /* TO BE FIXED */
          }
          gettimeofday (&curr_time, NULL);
          time = curr_time.tv_sec * 1000 + curr_time.tv_usec / 1000;
          job->script_start_time = time;
     }

     /* Check if there is a window is to grab.  */
     if ( !job->windowid ) {
          printf("No selected window to grab.\n");
     }

     process_next_token(job);

     /* Close and cleanup everything. */
     if ( job->script_state & SCRIPT_DONE ) {
          exit(0);
     } else {
          /* Record the parameters of the next capture. */
          static char file[PATH_MAX+1];
          unsigned long time;

          sprintf(file, job->file, job->pic_no);
          time = get_relative_time(job);
          fprintf(job->to_result, "image \"%s\" at %lu\n", file, time);
     }
     if ( job->script_state & SCRIPT_STOP ) {
          terminate_script(job);
     }

     /* Capture frame now */
}

/* end of script.c */
